typedef unsigned char opcode_t;
typedef unsigned char subcode_t;
typedef unsigned char ref_type_t;
typedef unsigned char addr_t;

typedef struct memory_ref_t {
  ref_type_t type;
  addr_t     addr;
} memory_ref_t;

/* CONSTANT, DIRECT, INDIRECT */
#define SET_MODE(type,value) ((type) = ((type)&0x0f)| (((value)&0x0f)<<4))
#define GET_MODE(type) (((type)>>4)&0x0f)

#define CONSTANT      0
#define DIRECT        1
#define INDIRECT      2

typedef struct {
  memory_ref_t arg;
} onearg_t;

typedef struct assign_t {
  memory_ref_t arg1;
  memory_ref_t arg2;
}  twoargs_t;

typedef struct if_t {
  memory_ref_t arg1;
  memory_ref_t arg2;
  memory_ref_t arg3;
} if_t;

typedef struct instruction_t {
  opcode_t opcode;
  union {
    twoargs_t twoargs;
    if_t      si;
    onearg_t  onearg;
  } datas;
  unsigned char padding; /* Should be zero */
} instruction_t;

#define UNK          -1
#define ASSIGN_PLUS  0
#define ASSIGN_MINUS 1
#define ASSIGN_MULT  2
#define ASSIGN       3
#define IF           4
#define JUMP         5
#define CALL         6
#define RET          7
#define FORK         8
#define EXEC         9
#define EXIT         10
#define WAIT         11
#define SIGNAL       12
#define KILL         13
#define READ         14
#define WRITE        15
#define GETPPID      16
#define LCK          17
#define UNLCK        18
#define SETPRIO      19
#define LAST_OPCODE  19

#define U_SIGILL     0
#define U_SIGMEM     1
#define U_SIGUSER    2
#define U_SIGKILL    3

static char *opcodes[] = {
  /* Standard instruction set */
  "ASSIGN+", "ASSIGN-", "ASSIGN*", "ASSIGN", "IF", "JUMP", "CALL", "RET",
  /* System call set */
  "FORK", "EXEC", "EXIT", "WAIT", "SIGNAL", "KILL", "READ", "WRITE",
  "GETPPID", "LCK", "UNLCK", "SETPRIO"
};
